@@ -28,7 +28,7 @@
 
 #include <haproxy/activity.h>
 #include <haproxy/api.h>
-#include <haproxy/applet-t.h>
+#include <haproxy/applet.h>
 #include <haproxy/base64.h>
 #include <haproxy/cfgparse.h>
 #include <haproxy/channel.h>
@@ -52,12 +52,13 @@
 #include <haproxy/protocol.h>
 #include <haproxy/proxy.h>
 #include <haproxy/sample-t.h>
+#include <haproxy/sc_strm.h>
 #include <haproxy/server.h>
 #include <haproxy/session.h>
 #include <haproxy/sock.h>
 #include <haproxy/stats-t.h>
+#include <haproxy/stconn.h>
 #include <haproxy/stream.h>
-#include <haproxy/stream_interface.h>
 #include <haproxy/task.h>
 #include <haproxy/ticks.h>
 #include <haproxy/time.h>
@@ -85,6 +86,24 @@ extern const char *stat_status_codes[];
 
 struct proxy *mworker_proxy; /* CLI proxy of the master */
 
+/* CLI context for the "show env" command */
+struct show_env_ctx {
+	char **var;      /* first variable to show */
+	int show_one;    /* stop after showing the first one */
+};
+
+/* CLI context for the "show fd" command */
+struct show_fd_ctx {
+	int fd;          /* first FD to show */
+	int show_one;    /* stop after showing one FD */
+};
+
+/* CLI context for the "show cli sockets" command */
+struct show_sock_ctx {
+	struct bind_conf *bind_conf;
+	struct listener *listener;
+};
+
 static int cmp_kw_entries(const void *a, const void *b)
 {
 	const struct cli_kw *l = *(const struct cli_kw **)a;
@@ -122,7 +141,8 @@ static char *cli_gen_usage_msg(struct appctx *appctx, char * const *args)
 		for (kw = &kw_list->kw[0]; kw->str_kw[0]; kw++) {
 			if (kw->level & ~appctx->cli_level & (ACCESS_MASTER_ONLY|ACCESS_EXPERT|ACCESS_EXPERIMENTAL))
 				continue;
-			if ((appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) ==
+			if (!(appctx->cli_level & ACCESS_MCLI_DEBUG) &&
+			    (appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) ==
 			    (ACCESS_MASTER_ONLY|ACCESS_MASTER))
 				continue;
 
@@ -162,8 +182,9 @@ static char *cli_gen_usage_msg(struct appctx *appctx, char * const *args)
 			for (kw = &kw_list->kw[0]; kw->str_kw[0]; kw++) {
 				if (kw->level & ~appctx->cli_level & (ACCESS_MASTER_ONLY|ACCESS_EXPERT|ACCESS_EXPERIMENTAL))
 					continue;
-				if ((appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) ==
-				    (ACCESS_MASTER_ONLY|ACCESS_MASTER))
+				if (!(appctx->cli_level & ACCESS_MCLI_DEBUG) &&
+				    ((appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) ==
+				    (ACCESS_MASTER_ONLY|ACCESS_MASTER)))
 					continue;
 
 				for (idx = 0; idx < length; idx++) {
@@ -249,11 +270,13 @@ static char *cli_gen_usage_msg(struct appctx *appctx, char * const *args)
 			if (kw->level & ~appctx->cli_level & (ACCESS_MASTER_ONLY|ACCESS_EXPERT|ACCESS_EXPERIMENTAL))
 				continue;
 
-			/* in master don't display commands that have neither the master bit
-			 * nor the master-only bit.
+			/* in master, if the CLI don't have the
+			 * ACCESS_MCLI_DEBUG don't display commands that have
+			 * neither the master bit nor the master-only bit.
 			 */
-			if ((appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) ==
-			    (ACCESS_MASTER_ONLY|ACCESS_MASTER))
+			if (!(appctx->cli_level & ACCESS_MCLI_DEBUG) &&
+			    ((appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) ==
+			    (ACCESS_MASTER_ONLY|ACCESS_MASTER)))
 				continue;
 
 			for (idx = 0; idx < length; idx++) {
@@ -285,10 +308,7 @@ static char *cli_gen_usage_msg(struct appctx *appctx, char * const *args)
 	chunk_dup(&out, tmp);
 	dynamic_usage_msg = out.area;
 
-	appctx->ctx.cli.severity = LOG_INFO;
-	appctx->ctx.cli.msg = dynamic_usage_msg;
-	appctx->st0 = CLI_ST_PRINT;
-
+	cli_msg(appctx, LOG_INFO, dynamic_usage_msg);
 	return dynamic_usage_msg;
 }
 
@@ -364,6 +384,43 @@ void cli_register_kw(struct cli_kw_list *kw_list)
 	LIST_APPEND(&cli_keywords.list, &kw_list->list);
 }
 
+/* list all known keywords on stdout, one per line */
+void cli_list_keywords(void)
+{
+	struct cli_kw_list *kw_list;
+	struct cli_kw *kwp, *kwn, *kw;
+	int idx;
+
+	for (kwn = kwp = NULL;; kwp = kwn) {
+		list_for_each_entry(kw_list, &cli_keywords.list, list) {
+			/* note: we sort based on the usage message when available,
+			 * otherwise we fall back to the first keyword.
+			 */
+			for (kw = &kw_list->kw[0]; kw->str_kw[0]; kw++) {
+				if (strordered(kwp ? kwp->usage ? kwp->usage : kwp->str_kw[0] : NULL,
+					       kw->usage ? kw->usage : kw->str_kw[0],
+					       kwn != kwp ? kwn->usage ? kwn->usage : kwn->str_kw[0] : NULL))
+					kwn = kw;
+			}
+		}
+
+		if (kwn == kwp)
+			break;
+
+		for (idx = 0; kwn->str_kw[idx]; idx++) {
+			printf("%s ", kwn->str_kw[idx]);
+		}
+		if (kwn->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER))
+			printf("[MASTER] ");
+		if (!(kwn->level & ACCESS_MASTER_ONLY))
+			printf("[WORKER] ");
+		if (kwn->level & ACCESS_EXPERT)
+			printf("[EXPERT] ");
+		if (kwn->level & ACCESS_EXPERIMENTAL)
+			printf("[EXPERIM] ");
+		printf("\n");
+	}
+}
 
 /* allocate a new stats frontend named <name>, and return it
  * (or NULL in case of lack of memory).
@@ -665,7 +722,7 @@ static int cli_get_severity_output(struct appctx *appctx)
 {
 	if (appctx->cli_severity_output)
 		return appctx->cli_severity_output;
-	return strm_li(si_strm(appctx->owner))->bind_conf->severity_output;
+	return strm_li(appctx_strm(appctx))->bind_conf->severity_output;
 }
 
 /* Processes the CLI interpreter on the stats socket. This function is called
@@ -685,8 +742,15 @@ static int cli_parse_request(struct appctx *appctx)
 	int i = 0;
 	struct cli_kw *kw;
 
-	appctx->st2 = 0;
-	memset(&appctx->ctx.cli, 0, sizeof(appctx->ctx.cli));
+	appctx->_st2 = 0;
+
+	/* temporary for 2.6: let's make sure we clean the whole shared
+	 * context.
+	 */
+	if (sizeof(appctx->ctx) > sizeof(appctx->svc))
+		memset(&appctx->ctx, 0, sizeof(appctx->ctx));
+	else
+		memset(&appctx->svc, 0, sizeof(appctx->svc));
 
 	p = appctx->chunk->area;
 	end = p + appctx->chunk->data;
@@ -750,7 +814,8 @@ static int cli_parse_request(struct appctx *appctx)
 	kw = cli_find_kw(args);
 	if (!kw ||
 	    (kw->level & ~appctx->cli_level & ACCESS_MASTER_ONLY) ||
-	    (appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) == (ACCESS_MASTER_ONLY|ACCESS_MASTER)) {
+	    (!(appctx->cli_level & ACCESS_MCLI_DEBUG) &&
+	     (appctx->cli_level & ~kw->level & (ACCESS_MASTER_ONLY|ACCESS_MASTER)) == (ACCESS_MASTER_ONLY|ACCESS_MASTER))) {
 		/* keyword not found in this mode */
 		cli_gen_usage_msg(appctx, args);
 		return 0;
@@ -822,7 +887,7 @@ static int cli_output_msg(struct channel *chn, const char *msg, int severity, in
 	return ci_putblk(chn, tmp->area, strlen(tmp->area));
 }
 
-/* This I/O handler runs as an applet embedded in a stream interface. It is
+/* This I/O handler runs as an applet embedded in a stream connector. It is
  * used to processes I/O from/to the stats unix socket. The system relies on a
  * state machine handling requests and various responses. We read a request,
  * then we process it and send the response, and we possibly display a prompt.
@@ -832,14 +897,14 @@ static int cli_output_msg(struct channel *chn, const char *msg, int severity, in
  */
 static void cli_io_handler(struct appctx *appctx)
 {
-	struct stream_interface *si = appctx->owner;
-	struct channel *req = si_oc(si);
-	struct channel *res = si_ic(si);
-	struct bind_conf *bind_conf = strm_li(si_strm(si))->bind_conf;
+	struct stconn *sc = appctx_sc(appctx);
+	struct channel *req = sc_oc(sc);
+	struct channel *res = sc_ic(sc);
+	struct bind_conf *bind_conf = strm_li(__sc_strm(sc))->bind_conf;
 	int reql;
 	int len;
 
-	if (unlikely(si->state == SI_ST_DIS || si->state == SI_ST_CLO))
+	if (unlikely(sc->state == SC_ST_DIS || sc->state == SC_ST_CLO))
 		goto out;
 
 	/* Check if the input buffer is available. */
@@ -852,8 +917,8 @@ static void cli_io_handler(struct appctx *appctx)
 
 	while (1) {
 		if (appctx->st0 == CLI_ST_INIT) {
-			/* Stats output not initialized yet */
-			memset(&appctx->ctx.stats, 0, sizeof(appctx->ctx.stats));
+			/* CLI/stats not initialized yet */
+			memset(&appctx->ctx, 0, sizeof(appctx->ctx));
 			/* reset severity to default at init */
 			appctx->cli_severity_output = bind_conf->severity_output;
 			appctx->st0 = CLI_ST_GETREQ;
@@ -863,7 +928,7 @@ static void cli_io_handler(struct appctx *appctx)
 			/* Let's close for real now. We just close the request
 			 * side, the conditions below will complete if needed.
 			 */
-			si_shutw(si);
+			sc_shutw(sc);
 			free_trash_chunk(appctx->chunk);
 			appctx->chunk = NULL;
 			break;
@@ -885,8 +950,8 @@ static void cli_io_handler(struct appctx *appctx)
 			/* ensure we have some output room left in the event we
 			 * would want to return some info right after parsing.
 			 */
-			if (buffer_almost_full(si_ib(si))) {
-				si_rx_room_blk(si);
+			if (buffer_almost_full(sc_ib(sc))) {
+				sc_need_room(sc);
 				break;
 			}
 
@@ -897,10 +962,10 @@ static void cli_io_handler(struct appctx *appctx)
 			 */
 
 			if (appctx->st1 & APPCTX_CLI_ST1_PAYLOAD)
-				reql = co_getline(si_oc(si), str,
+				reql = co_getline(sc_oc(sc), str,
 				                  appctx->chunk->size - appctx->chunk->data - 1);
 			else
-				reql = co_getdelim(si_oc(si), str,
+				reql = co_getdelim(sc_oc(sc), str,
 				                   appctx->chunk->size - appctx->chunk->data - 1,
 				                   "\n;", '\\');
 
@@ -985,10 +1050,11 @@ static void cli_io_handler(struct appctx *appctx)
 			}
 
 			/* re-adjust req buffer */
-			co_skip(si_oc(si), reql);
+			co_skip(sc_oc(sc), reql);
 			req->flags |= CF_READ_DONTWAIT; /* we plan to read small requests */
 		}
 		else {	/* output functions */
+			struct cli_print_ctx *ctx;
 			const char *msg;
 			int sev;
 
@@ -999,15 +1065,17 @@ static void cli_io_handler(struct appctx *appctx)
 			case CLI_ST_PRINT_ERR:   /* print const error in msg */
 			case CLI_ST_PRINT_DYN:   /* print dyn message in msg, free */
 			case CLI_ST_PRINT_FREE:  /* print dyn error in err, free */
+				/* the message is in the svcctx */
+				ctx = applet_reserve_svcctx(appctx, sizeof(*ctx));
 				if (appctx->st0 == CLI_ST_PRINT || appctx->st0 == CLI_ST_PRINT_ERR) {
 					sev = appctx->st0 == CLI_ST_PRINT_ERR ?
-						LOG_ERR : appctx->ctx.cli.severity;
-					msg = appctx->ctx.cli.msg;
+						LOG_ERR : ctx->severity;
+					msg = ctx->msg;
 				}
 				else if (appctx->st0 == CLI_ST_PRINT_DYN || appctx->st0 == CLI_ST_PRINT_FREE) {
 					sev = appctx->st0 == CLI_ST_PRINT_FREE ?
-						LOG_ERR : appctx->ctx.cli.severity;
-					msg = appctx->ctx.cli.err;
+						LOG_ERR : ctx->severity;
+					msg = ctx->err;
 					if (!msg) {
 						sev = LOG_ERR;
 						msg = "Out of memory.\n";
@@ -1021,12 +1089,12 @@ static void cli_io_handler(struct appctx *appctx)
 				if (cli_output_msg(res, msg, sev, cli_get_severity_output(appctx)) != -1) {
 					if (appctx->st0 == CLI_ST_PRINT_FREE ||
 					    appctx->st0 == CLI_ST_PRINT_DYN) {
-						ha_free(&appctx->ctx.cli.err);
+						ha_free(&ctx->err);
 					}
 					appctx->st0 = CLI_ST_PROMPT;
 				}
 				else
-					si_rx_room_blk(si);
+					sc_need_room(sc);
 				break;
 
 			case CLI_ST_CALLBACK: /* use custom pointer */
@@ -1040,7 +1108,7 @@ static void cli_io_handler(struct appctx *appctx)
 					}
 				break;
 			default: /* abnormal state */
-				si->flags |= SI_FL_ERR;
+				se_fl_set(appctx->sedesc, SE_FL_ERROR);
 				break;
 			}
 
@@ -1063,10 +1131,8 @@ static void cli_io_handler(struct appctx *appctx)
 						prompt = "\n";
 				}
 
-				if (ci_putstr(si_ic(si), prompt) != -1)
+				if (applet_putstr(appctx, prompt) != -1)
 					appctx->st0 = CLI_ST_GETREQ;
-				else
-					si_rx_room_blk(si);
 			}
 
 			/* If the output functions are still there, it means they require more room. */
@@ -1095,42 +1161,42 @@ static void cli_io_handler(struct appctx *appctx)
 			 * refills the buffer with new bytes in non-interactive
 			 * mode, avoiding to close on apparently empty commands.
 			 */
-			if (co_data(si_oc(si))) {
+			if (co_data(sc_oc(sc))) {
 				appctx_wakeup(appctx);
 				goto out;
 			}
 		}
 	}
 
-	if ((res->flags & CF_SHUTR) && (si->state == SI_ST_EST)) {
-		DPRINTF(stderr, "%s@%d: si to buf closed. req=%08x, res=%08x, st=%d\n",
-			__FUNCTION__, __LINE__, req->flags, res->flags, si->state);
+	if ((res->flags & CF_SHUTR) && (sc->state == SC_ST_EST)) {
+		DPRINTF(stderr, "%s@%d: sc to buf closed. req=%08x, res=%08x, st=%d\n",
+			__FUNCTION__, __LINE__, req->flags, res->flags, sc->state);
 		/* Other side has closed, let's abort if we have no more processing to do
 		 * and nothing more to consume. This is comparable to a broken pipe, so
 		 * we forward the close to the request side so that it flows upstream to
 		 * the client.
 		 */
-		si_shutw(si);
+		sc_shutw(sc);
 	}
 
-	if ((req->flags & CF_SHUTW) && (si->state == SI_ST_EST) && (appctx->st0 < CLI_ST_OUTPUT)) {
-		DPRINTF(stderr, "%s@%d: buf to si closed. req=%08x, res=%08x, st=%d\n",
-			__FUNCTION__, __LINE__, req->flags, res->flags, si->state);
+	if ((req->flags & CF_SHUTW) && (sc->state == SC_ST_EST) && (appctx->st0 < CLI_ST_OUTPUT)) {
+		DPRINTF(stderr, "%s@%d: buf to sc closed. req=%08x, res=%08x, st=%d\n",
+			__FUNCTION__, __LINE__, req->flags, res->flags, sc->state);
 		/* We have no more processing to do, and nothing more to send, and
 		 * the client side has closed. So we'll forward this state downstream
 		 * on the response buffer.
 		 */
-		si_shutr(si);
+		sc_shutr(sc);
 		res->flags |= CF_READ_NULL;
 	}
 
  out:
 	DPRINTF(stderr, "%s@%d: st=%d, rqf=%x, rpf=%x, rqh=%lu, rqs=%lu, rh=%lu, rs=%lu\n",
 		__FUNCTION__, __LINE__,
-		si->state, req->flags, res->flags, ci_data(req), co_data(req), ci_data(res), co_data(res));
+		sc->state, req->flags, res->flags, ci_data(req), co_data(req), ci_data(res), co_data(res));
 }
 
-/* This is called when the stream interface is closed. For instance, upon an
+/* This is called when the stream connector is closed. For instance, upon an
  * external abort, we won't call the i/o handler anymore so we may need to
  * remove back references to the stream currently being dumped.
  */
@@ -1144,21 +1210,24 @@ static void cli_release_handler(struct appctx *appctx)
 		appctx->io_release = NULL;
 	}
 	else if (appctx->st0 == CLI_ST_PRINT_FREE || appctx->st0 == CLI_ST_PRINT_DYN) {
-		ha_free(&appctx->ctx.cli.err);
+		struct cli_print_ctx *ctx = applet_reserve_svcctx(appctx, sizeof(*ctx));
+
+		ha_free(&ctx->err);
 	}
 }
 
 /* This function dumps all environmnent variables to the buffer. It returns 0
  * if the output buffer is full and it needs to be called again, otherwise
- * non-zero. Dumps only one entry if st2 == STAT_ST_END. It uses cli.p0 as the
- * pointer to the current variable.
+ * non-zero. It takes its context from the show_env_ctx in svcctx, and will
+ * start from ->var and dump only one variable if ->show_one is set.
  */
 static int cli_io_handler_show_env(struct appctx *appctx)
 {
-	struct stream_interface *si = appctx->owner;
-	char **var = appctx->ctx.cli.p0;
+	struct show_env_ctx *ctx = appctx->svcctx;
+	struct stconn *sc = appctx_sc(appctx);
+	char **var = ctx->var;
 
-	if (unlikely(si_ic(si)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
+	if (unlikely(sc_ic(sc)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
 		return 1;
 
 	chunk_reset(&trash);
@@ -1169,14 +1238,13 @@ static int cli_io_handler_show_env(struct appctx *appctx)
 	while (*var) {
 		chunk_printf(&trash, "%s\n", *var);
 
-		if (ci_putchk(si_ic(si), &trash) == -1) {
-			si_rx_room_blk(si);
+		if (applet_putchk(appctx, &trash) == -1)
 			return 0;
-		}
-		if (appctx->st2 == STAT_ST_END)
+
+		if (ctx->show_one)
 			break;
 		var++;
-		appctx->ctx.cli.p0 = var;
+		ctx->var = var;
 	}
 
 	/* dump complete */
@@ -1185,16 +1253,18 @@ static int cli_io_handler_show_env(struct appctx *appctx)
 
 /* This function dumps all file descriptors states (or the requested one) to
  * the buffer. It returns 0 if the output buffer is full and it needs to be
- * called again, otherwise non-zero. Dumps only one entry if st2 == STAT_ST_END.
- * It uses cli.i0 as the fd number to restart from.
+ * called again, otherwise non-zero. It takes its context from the show_fd_ctx
+ * in svcctx, only dumps one entry if ->show_one is non-zero, and (re)starts
+ * from ->fd.
  */
 static int cli_io_handler_show_fd(struct appctx *appctx)
 {
-	struct stream_interface *si = appctx->owner;
-	int fd = appctx->ctx.cli.i0;
+	struct stconn *sc = appctx_sc(appctx);
+	struct show_fd_ctx *fdctx = appctx->svcctx;
+	int fd = fdctx->fd;
 	int ret = 1;
 
-	if (unlikely(si_ic(si)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
+	if (unlikely(sc_ic(sc)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
 		goto end;
 
 	chunk_reset(&trash);
@@ -1361,14 +1431,13 @@ static int cli_io_handler_show_fd(struct appctx *appctx)
 #endif
 		chunk_appendf(&trash, "%s\n", suspicious ? " !" : "");
 
-		if (ci_putchk(si_ic(si), &trash) == -1) {
-			si_rx_room_blk(si);
-			appctx->ctx.cli.i0 = fd;
+		if (applet_putchk(appctx, &trash) == -1) {
+			fdctx->fd = fd;
 			ret = 0;
 			break;
 		}
 	skip:
-		if (appctx->st2 == STAT_ST_END)
+		if (fdctx->show_one)
 			break;
 
 		fd++;
@@ -1389,10 +1458,10 @@ static int cli_io_handler_show_fd(struct appctx *appctx)
  */
 static int cli_io_handler_show_activity(struct appctx *appctx)
 {
-	struct stream_interface *si = appctx->owner;
+	struct stconn *sc = appctx_sc(appctx);
 	int thr;
 
-	if (unlikely(si_ic(si)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
+	if (unlikely(sc_ic(sc)->flags & (CF_WRITE_ERROR|CF_SHUTW)))
 		return 1;
 
 	chunk_reset(&trash);
@@ -1473,10 +1542,9 @@ static int cli_io_handler_show_activity(struct appctx *appctx)
 	chunk_appendf(&trash, "ctr2:");         SHOW_TOT(thr, activity[thr].ctr2);
 #endif
 
-	if (ci_putchk(si_ic(si), &trash) == -1) {
+	if (applet_putchk(appctx, &trash) == -1) {
 		chunk_reset(&trash);
 		chunk_printf(&trash, "[output too large, cannot dump]\n");
-		si_rx_room_blk(si);
 	}
 
 #undef SHOW_AVG
@@ -1487,97 +1555,91 @@ static int cli_io_handler_show_activity(struct appctx *appctx)
 
 /*
  * CLI IO handler for `show cli sockets`.
- * Uses ctx.cli.p0 to store the bind_conf pointer, and cli.p1 for the listener.
+ * Uses the svcctx as a show_sock_ctx to store/retrieve the bind_conf and the
+ * listener pointers.
  */
 static int cli_io_handler_show_cli_sock(struct appctx *appctx)
 {
-	struct bind_conf *bind_conf = appctx->ctx.cli.p0;
-	struct stream_interface *si = appctx->owner;
+	struct show_sock_ctx *ctx = applet_reserve_svcctx(appctx, sizeof(*ctx));
+	struct bind_conf *bind_conf = ctx->bind_conf;
+
+	if (!global.cli_fe)
+		goto done;
 
 	chunk_reset(&trash);
 
-	switch (appctx->st2) {
-		case STAT_ST_INIT:
-			chunk_printf(&trash, "# socket lvl processes\n");
-			if (ci_putchk(si_ic(si), &trash) == -1) {
-				si_rx_room_blk(si);
-				return 0;
-			}
-			appctx->st2 = STAT_ST_LIST;
-			/* fall through */
-
-		case STAT_ST_LIST:
-			if (global.cli_fe) {
-				if (!bind_conf)
-					bind_conf = LIST_ELEM(global.cli_fe->conf.bind.n, typeof(bind_conf), by_fe);
-
-				list_for_each_entry_from(bind_conf, &global.cli_fe->conf.bind, by_fe) {
-					struct listener *l = appctx->ctx.cli.p1;
-
-					if (!l)
-						l = LIST_ELEM(bind_conf->listeners.n, typeof(l), by_bind);
-
-					list_for_each_entry_from(l, &bind_conf->listeners, by_bind) {
-						char addr[46];
-						char port[6];
-
-						if (l->rx.addr.ss_family == AF_UNIX) {
-							const struct sockaddr_un *un;
-
-							un = (struct sockaddr_un *)&l->rx.addr;
-							if (un->sun_path[0] == '\0') {
-								chunk_appendf(&trash, "abns@%s ", un->sun_path+1);
-							} else {
-								chunk_appendf(&trash, "unix@%s ", un->sun_path);
-							}
-						} else if (l->rx.addr.ss_family == AF_INET) {
-							addr_to_str(&l->rx.addr, addr, sizeof(addr));
-							port_to_str(&l->rx.addr, port, sizeof(port));
-							chunk_appendf(&trash, "ipv4@%s:%s ", addr, port);
-						} else if (l->rx.addr.ss_family == AF_INET6) {
-							addr_to_str(&l->rx.addr, addr, sizeof(addr));
-							port_to_str(&l->rx.addr, port, sizeof(port));
-							chunk_appendf(&trash, "ipv6@[%s]:%s ", addr, port);
-						} else if (l->rx.addr.ss_family == AF_CUST_SOCKPAIR) {
-							chunk_appendf(&trash, "sockpair@%d ", ((struct sockaddr_in *)&l->rx.addr)->sin_addr.s_addr);
-						} else
-							chunk_appendf(&trash, "unknown ");
-
-						if ((bind_conf->level & ACCESS_LVL_MASK) == ACCESS_LVL_ADMIN)
-							chunk_appendf(&trash, "admin ");
-						else if ((bind_conf->level & ACCESS_LVL_MASK) == ACCESS_LVL_OPER)
-							chunk_appendf(&trash, "operator ");
-						else if ((bind_conf->level & ACCESS_LVL_MASK) == ACCESS_LVL_USER)
-							chunk_appendf(&trash, "user ");
-						else
-							chunk_appendf(&trash, "  ");
-
-						chunk_appendf(&trash, "all\n");
-
-						if (ci_putchk(si_ic(si), &trash) == -1) {
-							/* buffer full, we must yield */
-							appctx->ctx.cli.p0 = bind_conf;
-							appctx->ctx.cli.p1 = l;
-							si_rx_room_blk(si);
-							return 0;
-						}
-					}
-				}
-			}
-			/* fall through */
-		default:
-			appctx->st2 = STAT_ST_FIN;
-			return 1;
+	if (!bind_conf) {
+		/* first call */
+		if (applet_putstr(appctx, "# socket lvl processes\n") == -1)
+			goto full;
+		bind_conf = LIST_ELEM(global.cli_fe->conf.bind.n, typeof(bind_conf), by_fe);
 	}
+
+	list_for_each_entry_from(bind_conf, &global.cli_fe->conf.bind, by_fe) {
+		struct listener *l = ctx->listener;
+
+		if (!l)
+			l = LIST_ELEM(bind_conf->listeners.n, typeof(l), by_bind);
+
+		list_for_each_entry_from(l, &bind_conf->listeners, by_bind) {
+			char addr[46];
+			char port[6];
+
+			if (l->rx.addr.ss_family == AF_UNIX) {
+				const struct sockaddr_un *un;
+
+				un = (struct sockaddr_un *)&l->rx.addr;
+				if (un->sun_path[0] == '\0') {
+					chunk_appendf(&trash, "abns@%s ", un->sun_path+1);
+				} else {
+					chunk_appendf(&trash, "unix@%s ", un->sun_path);
+				}
+			} else if (l->rx.addr.ss_family == AF_INET) {
+				addr_to_str(&l->rx.addr, addr, sizeof(addr));
+				port_to_str(&l->rx.addr, port, sizeof(port));
+				chunk_appendf(&trash, "ipv4@%s:%s ", addr, port);
+			} else if (l->rx.addr.ss_family == AF_INET6) {
+				addr_to_str(&l->rx.addr, addr, sizeof(addr));
+				port_to_str(&l->rx.addr, port, sizeof(port));
+				chunk_appendf(&trash, "ipv6@[%s]:%s ", addr, port);
+			} else if (l->rx.addr.ss_family == AF_CUST_SOCKPAIR) {
+				chunk_appendf(&trash, "sockpair@%d ", ((struct sockaddr_in *)&l->rx.addr)->sin_addr.s_addr);
+			} else
+				chunk_appendf(&trash, "unknown ");
+
+			if ((bind_conf->level & ACCESS_LVL_MASK) == ACCESS_LVL_ADMIN)
+				chunk_appendf(&trash, "admin ");
+			else if ((bind_conf->level & ACCESS_LVL_MASK) == ACCESS_LVL_OPER)
+				chunk_appendf(&trash, "operator ");
+			else if ((bind_conf->level & ACCESS_LVL_MASK) == ACCESS_LVL_USER)
+				chunk_appendf(&trash, "user ");
+			else
+				chunk_appendf(&trash, "  ");
+
+			chunk_appendf(&trash, "all\n");
+
+			if (applet_putchk(appctx, &trash) == -1) {
+				ctx->bind_conf = bind_conf;
+				ctx->listener  = l;
+				goto full;
+			}
+		}
+	}
+ done:
+	return 1;
+ full:
+	return 0;
 }
 
 
 /* parse a "show env" CLI request. Returns 0 if it needs to continue, 1 if it
- * wants to stop here. It puts the variable to be dumped into cli.p0 if a single
- * variable is requested otherwise puts environ there.
+ * wants to stop here. It reserves a sohw_env_ctx where it puts the variable to
+ * be dumped as well as a flag if a single variable is requested, otherwise puts
+ * environ there.
  */
 static int cli_parse_show_env(char **args, char *payload, struct appctx *appctx, void *private)
 {
+	struct show_env_ctx *ctx = applet_reserve_svcctx(appctx, sizeof(*ctx));
 	extern char **environ;
 	char **var;
 
@@ -1597,26 +1659,26 @@ static int cli_parse_show_env(char **args, char *payload, struct appctx *appctx,
 		if (!*var)
 			return cli_err(appctx, "Variable not found\n");
 
-		appctx->st2 = STAT_ST_END;
+		ctx->show_one = 1;
 	}
-	appctx->ctx.cli.p0 = var;
+	ctx->var = var;
 	return 0;
 }
 
 /* parse a "show fd" CLI request. Returns 0 if it needs to continue, 1 if it
- * wants to stop here. It puts the FD number into cli.i0 if a specific FD is
- * requested and sets st2 to STAT_ST_END, otherwise leaves 0 in i0.
+ * wants to stop here. It sets a show_fd_ctx context where, if a specific fd is
+ * requested, it puts the FD number into ->fd and sets ->show_one to 1.
  */
 static int cli_parse_show_fd(char **args, char *payload, struct appctx *appctx, void *private)
 {
+	struct show_fd_ctx *ctx = applet_reserve_svcctx(appctx, sizeof(*ctx));
+
 	if (!cli_has_level(appctx, ACCESS_LVL_OPER))
 		return 1;
 
-	appctx->ctx.cli.i0 = 0;
-
 	if (*args[2]) {
-		appctx->ctx.cli.i0 = atoi(args[2]);
-		appctx->st2 = STAT_ST_END;
+		ctx->fd = atoi(args[2]);
+		ctx->show_one = 1;
 	}
 	return 0;
 }
@@ -1624,8 +1686,7 @@ static int cli_parse_show_fd(char **args, char *payload, struct appctx *appctx,
 /* parse a "set timeout" CLI request. It always returns 1. */
 static int cli_parse_set_timeout(char **args, char *payload, struct appctx *appctx, void *private)
 {
-	struct stream_interface *si = appctx->owner;
-	struct stream *s = si_strm(si);
+	struct stream *s = appctx_strm(appctx);
 
 	if (strcmp(args[2], "cli") == 0) {
 		unsigned timeout;
@@ -1746,6 +1807,10 @@ static int cli_parse_expert_experimental_mode(char **args, char *payload, struct
 	char *level_str;
 	char *output = NULL;
 
+	/* this will ask the applet to not output a \n after the command */
+	if (*args[1] && *args[2] && strcmp(args[2], "-") == 0)
+		appctx->st1 |= APPCTX_CLI_ST1_NOLF;
+
 	if (!cli_has_level(appctx, ACCESS_LVL_ADMIN))
 		return 1;
 
@@ -1757,6 +1822,10 @@ static int cli_parse_expert_experimental_mode(char **args, char *payload, struct
 		level = ACCESS_EXPERIMENTAL;
 		level_str = "experimental-mode";
 	}
+	else if (strcmp(args[0], "mcli-debug-mode") == 0) {
+		level = ACCESS_MCLI_DEBUG;
+		level_str = "mcli-debug-mode";
+	}
 	else {
 		return 1;
 	}
@@ -1900,9 +1969,9 @@ static int _getsocks(char **args, char *payload, struct appctx *appctx, void *pr
 	char *cmsgbuf = NULL;
 	unsigned char *tmpbuf = NULL;
 	struct cmsghdr *cmsg;
-	struct stream_interface *si = appctx->owner;
-	struct stream *s = si_strm(si);
-	struct connection *remote = cs_conn(objt_cs(si_opposite(si)->end));
+	struct stconn *sc = appctx_sc(appctx);
+	struct stream *s = __sc_strm(sc);
+	struct connection *remote = sc_conn(sc_opposite(sc));
 	struct msghdr msghdr;
 	struct iovec iov;
 	struct timeval tv = { .tv_sec = 1, .tv_usec = 0 };
@@ -2100,7 +2169,7 @@ static int cli_parse_simple(char **args, char *payload, struct appctx *appctx, v
 void pcli_write_prompt(struct stream *s)
 {
 	struct buffer *msg = get_trash_chunk();
-	struct channel *oc = si_oc(&s->si[0]);
+	struct channel *oc = sc_oc(s->scf);
 
 	if (!(s->pcli_flags & PCLI_F_PROMPT))
 		return;
@@ -2109,10 +2178,29 @@ void pcli_write_prompt(struct stream *s)
 		chunk_appendf(msg, "+ ");
 	} else {
 		if (s->pcli_next_pid == 0)
-			chunk_appendf(msg, "master%s> ",
+			chunk_appendf(msg, "master%s",
 			              (proc_self->failedreloads > 0) ? "[ReloadFailed]" : "");
 		else
-			chunk_appendf(msg, "%d> ", s->pcli_next_pid);
+			chunk_appendf(msg, "%d", s->pcli_next_pid);
+
+		if (s->pcli_flags & (ACCESS_EXPERIMENTAL|ACCESS_EXPERT|ACCESS_MCLI_DEBUG)) {
+			chunk_appendf(msg, "(");
+
+			if (s->pcli_flags & ACCESS_EXPERIMENTAL)
+				chunk_appendf(msg, "x");
+
+			if (s->pcli_flags & ACCESS_EXPERT)
+				chunk_appendf(msg, "e");
+
+			if (s->pcli_flags & ACCESS_MCLI_DEBUG)
+				chunk_appendf(msg, "d");
+
+			chunk_appendf(msg, ")");
+		}
+
+		chunk_appendf(msg, "> ");
+
+
 	}
 	co_inject(oc, msg->area, msg->data);
 }
@@ -2125,7 +2213,7 @@ void pcli_reply_and_close(struct stream *s, const char *msg)
 	struct buffer *buf = get_trash_chunk();
 
 	chunk_initstr(buf, msg);
-	si_retnclose(&s->si[0], buf);
+	stream_retnclose(s, buf);
 }
 
 static enum obj_type *pcli_pid_to_server(int proc_pid)
@@ -2257,6 +2345,36 @@ int pcli_find_and_exec_kw(struct stream *s, char **args, int argl, char **errmsg
 		s->pcli_flags &= ~ACCESS_LVL_MASK;
 		s->pcli_flags |= ACCESS_LVL_USER;
 		return argl;
+
+	} else if (strcmp(args[0], "expert-mode") == 0) {
+		if (!pcli_has_level(s, ACCESS_LVL_ADMIN)) {
+			memprintf(errmsg, "Permission denied!\n");
+			return -1;
+		}
+
+		s->pcli_flags &= ~ACCESS_EXPERT;
+		if ((argl > 1) && (strcmp(args[1], "on") == 0))
+			s->pcli_flags |= ACCESS_EXPERT;
+		return argl;
+
+	} else if (strcmp(args[0], "experimental-mode") == 0) {
+		if (!pcli_has_level(s, ACCESS_LVL_ADMIN)) {
+			memprintf(errmsg, "Permission denied!\n");
+			return -1;
+		}
+		s->pcli_flags &= ~ACCESS_EXPERIMENTAL;
+		if ((argl > 1) && (strcmp(args[1], "on") == 0))
+			s->pcli_flags |= ACCESS_EXPERIMENTAL;
+		return argl;
+	} else if (strcmp(args[0], "mcli-debug-mode") == 0) {
+		if (!pcli_has_level(s, ACCESS_LVL_ADMIN)) {
+			memprintf(errmsg, "Permission denied!\n");
+			return -1;
+		}
+		s->pcli_flags &= ~ACCESS_MCLI_DEBUG;
+		if ((argl > 1) && (strcmp(args[1], "on") == 0))
+			s->pcli_flags |= ACCESS_MCLI_DEBUG;
+		return argl;
 	}
 
 	return 0;
@@ -2372,7 +2490,7 @@ int pcli_parse_request(struct stream *s, struct channel *req, char **errmsg, int
 	wtrim = pcli_find_and_exec_kw(s, args, argl, errmsg, next_pid);
 
 	/* End of words are ending by \0, we need to replace the \0s by spaces
-1	   before forwarding them */
+	   before forwarding them */
 	p = str;
 	while (p < end-1) {
 		if (*p == '\0')
@@ -2405,6 +2523,21 @@ int pcli_parse_request(struct stream *s, struct channel *req, char **errmsg, int
 	}
 
 	if (ret > 1) {
+
+		/* the mcli-debug-mode is only sent to the applet of the master */
+		if ((s->pcli_flags & ACCESS_MCLI_DEBUG) && *next_pid <= 0) {
+			ci_insert_line2(req, 0, "mcli-debug-mode on -", strlen("mcli-debug-mode on -"));
+			ret += strlen("mcli-debug-mode on -") + 2;
+		}
+		if (s->pcli_flags & ACCESS_EXPERIMENTAL) {
+			ci_insert_line2(req, 0, "experimental-mode on -", strlen("experimental-mode on -"));
+			ret += strlen("experimental-mode on -") + 2;
+		}
+		if (s->pcli_flags & ACCESS_EXPERT) {
+			ci_insert_line2(req, 0, "expert-mode on -", strlen("expert-mode on -"));
+			ret += strlen("expert-mode on -") + 2;
+		}
+
 		if (pcli_has_level(s, ACCESS_LVL_ADMIN)) {
 			goto end;
 		} else if (pcli_has_level(s, ACCESS_LVL_OPER)) {
@@ -2426,7 +2559,7 @@ int pcli_wait_for_request(struct stream *s, struct channel *req, int an_bit)
 	int to_forward;
 	char *errmsg = NULL;
 
-	/* Don't read the next command if still processing the reponse of the
+	/* Don't read the next command if still processing the response of the
 	 * current one. Just wait. At this stage, errors should be handled by
 	 * the response analyzer.
 	 */
@@ -2564,9 +2697,9 @@ int pcli_wait_for_response(struct stream *s, struct channel *rep, int an_bit)
 
 		pcli_write_prompt(s);
 
-		s->si[1].flags |= SI_FL_NOLINGER | SI_FL_NOHALF;
-		si_shutr(&s->si[1]);
-		si_shutw(&s->si[1]);
+		s->scb->flags |= SC_FL_NOLINGER | SC_FL_NOHALF;
+		sc_shutr(s->scb);
+		sc_shutw(s->scb);
 
 		/*
 		 * starting from there this the same code as
@@ -2633,23 +2766,30 @@ int pcli_wait_for_response(struct stream *s, struct channel *rep, int an_bit)
 		/* only release our endpoint if we don't intend to reuse the
 		 * connection.
 		 */
-		if (!si_conn_ready(&s->si[1])) {
-			si_release_endpoint(&s->si[1]);
+		if (!sc_conn_ready(s->scb)) {
 			s->srv_conn = NULL;
+			if (sc_reset_endp(s->scb) < 0) {
+				if (!s->conn_err_type)
+					s->conn_err_type = STRM_ET_CONN_OTHER;
+				if (s->srv_error)
+					s->srv_error(s, s->scb);
+				return 1;
+			}
+			se_fl_clr(s->scb->sedesc, ~SE_FL_DETACHED);
 		}
 
-		sockaddr_free(&s->si[1].dst);
+		sockaddr_free(&s->scb->dst);
 
-		s->si[1].state     = s->si[1].prev_state = SI_ST_INI;
-		s->si[1].err_type  = SI_ET_NONE;
-		s->si[1].conn_retries = 0;  /* used for logging too */
-		s->si[1].exp       = TICK_ETERNITY;
-		s->si[1].flags    &= SI_FL_ISBACK | SI_FL_DONT_WAKE; /* we're in the context of process_stream */
+		sc_set_state(s->scb, SC_ST_INI);
+		s->scb->flags &= SC_FL_ISBACK | SC_FL_DONT_WAKE; /* we're in the context of process_stream */
 		s->req.flags &= ~(CF_SHUTW|CF_SHUTW_NOW|CF_AUTO_CONNECT|CF_WRITE_ERROR|CF_STREAMER|CF_STREAMER_FAST|CF_NEVER_WAIT|CF_WROTE_DATA);
 		s->res.flags &= ~(CF_SHUTR|CF_SHUTR_NOW|CF_READ_ATTACHED|CF_READ_ERROR|CF_READ_NOEXP|CF_STREAMER|CF_STREAMER_FAST|CF_WRITE_PARTIAL|CF_NEVER_WAIT|CF_WROTE_DATA|CF_READ_NULL);
-		s->flags &= ~(SF_DIRECT|SF_ASSIGNED|SF_ADDR_SET|SF_BE_ASSIGNED|SF_FORCE_PRST|SF_IGNORE_PRST);
+		s->flags &= ~(SF_DIRECT|SF_ASSIGNED|SF_BE_ASSIGNED|SF_FORCE_PRST|SF_IGNORE_PRST);
 		s->flags &= ~(SF_CURR_SESS|SF_REDIRECTABLE|SF_SRV_REUSED);
 		s->flags &= ~(SF_ERR_MASK|SF_FINST_MASK|SF_REDISP);
+		s->conn_retries = 0;  /* used for logging too */
+		s->conn_exp = TICK_ETERNITY;
+		s->conn_err_type = STRM_ET_NONE;
 		/* reinitialise the current rule list pointer to NULL. We are sure that
 		 * any rulelist match the NULL pointer.
 		 */
@@ -2696,7 +2836,7 @@ int pcli_wait_for_response(struct stream *s, struct channel *rep, int an_bit)
 		s->res.rex = TICK_ETERNITY;
 		s->res.wex = TICK_ETERNITY;
 		s->res.analyse_exp = TICK_ETERNITY;
-		s->si[1].hcto = TICK_ETERNITY;
+		s->scb->hcto = TICK_ETERNITY;
 
 		/* we're removing the analysers, we MUST re-enable events detection.
 		 * We don't enable close on the response channel since it's either
@@ -2953,6 +3093,7 @@ int mworker_cli_sockpair_new(struct mworker_proc *mworker_proc, int proc)
 
 	bind_conf->level &= ~ACCESS_LVL_MASK;
 	bind_conf->level |= ACCESS_LVL_ADMIN; /* TODO: need to lower the rights with a CLI keyword*/
+	bind_conf->level |= ACCESS_FD_LISTENERS;
 
 	if (!memprintf(&path, "sockpair@%d", mworker_proc->ipc_fd[1])) {
 		ha_alert("Cannot allocate listener.\n");
@@ -3008,8 +3149,9 @@ static struct cli_kw_list cli_kws = {{ },{
 	{ { "prompt", NULL },                    NULL,                                                                                                cli_parse_simple, NULL, NULL, NULL, ACCESS_MASTER },
 	{ { "quit", NULL },                      NULL,                                                                                                cli_parse_simple, NULL, NULL, NULL, ACCESS_MASTER },
 	{ { "_getsocks", NULL },                 NULL,                                                                                                _getsocks, NULL },
-	{ { "expert-mode", NULL },               NULL,                                                                                                cli_parse_expert_experimental_mode, NULL }, // not listed
-	{ { "experimental-mode", NULL },         NULL,                                                                                                cli_parse_expert_experimental_mode, NULL }, // not listed
+	{ { "expert-mode", NULL },               NULL,                                                                                                cli_parse_expert_experimental_mode, NULL, NULL, NULL, ACCESS_MASTER }, // not listed
+	{ { "experimental-mode", NULL },         NULL,                                                                                                cli_parse_expert_experimental_mode, NULL, NULL, NULL, ACCESS_MASTER }, // not listed
+	{ { "mcli-debug-mode", NULL },         NULL,                                                                                                  cli_parse_expert_experimental_mode, NULL, NULL, NULL, ACCESS_MASTER_ONLY }, // not listed
 	{ { "set", "maxconn", "global",  NULL }, "set maxconn global <value>              : change the per-process maxconn setting",                  cli_parse_set_maxconn_global, NULL },
 	{ { "set", "rate-limit", NULL },         "set rate-limit <setting> <value>        : change a rate limiting value",                            cli_parse_set_ratelimit, NULL },
 	{ { "set", "severity-output",  NULL },   "set severity-output [none|number|string]: set presence of severity level in feedback information",  cli_parse_set_severity_output, NULL, NULL },
